@@ -4,6 +4,7 @@
 // not use this file except in compliance with the License. You may obtain
 // a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
 
+//go:build cse
 // +build cse
 
 package integration
@@ -16,8 +17,8 @@ import (
 	"go.mongodb.org/mongo-driver/bson/bsontype"
 	"go.mongodb.org/mongo-driver/bson/primitive"
 	"go.mongodb.org/mongo-driver/event"
-	"go.mongodb.org/mongo-driver/internal/testutil"
-	"go.mongodb.org/mongo-driver/internal/testutil/assert"
+	"go.mongodb.org/mongo-driver/internal/assert"
+	"go.mongodb.org/mongo-driver/internal/integtest"
 	"go.mongodb.org/mongo-driver/mongo"
 	"go.mongodb.org/mongo-driver/mongo/integration/mtest"
 	"go.mongodb.org/mongo-driver/mongo/options"
@@ -36,17 +37,17 @@ func createDataKeyAndEncrypt(mt *mtest.T, keyName string) primitive.Binary {
 		SetReadConcern(mtest.MajorityRc).
 		SetWriteConcern(mtest.MajorityWc)
 
-	testutil.AddTestServerAPIVersion(kvClientOpts)
+	integtest.AddTestServerAPIVersion(kvClientOpts)
 
 	kmsProvidersMap := map[string]map[string]interface{}{
 		"local": {"key": localMasterKey},
 	}
 
-	kvClient, err := mongo.Connect(mtest.Background, kvClientOpts)
-	defer kvClient.Disconnect(mtest.Background)
+	kvClient, err := mongo.Connect(context.Background(), kvClientOpts)
+	defer kvClient.Disconnect(context.Background())
 	assert.Nil(mt, err, "Connect error: %v", err)
 
-	err = kvClient.Database("keyvault").Collection("datakeys").Drop(mtest.Background)
+	err = kvClient.Database("keyvault").Collection("datakeys").Drop(context.Background())
 	assert.Nil(mt, err, "Drop error: %v", err)
 
 	ceOpts := options.ClientEncryption().
@@ -57,7 +58,7 @@ func createDataKeyAndEncrypt(mt *mtest.T, keyName string) primitive.Binary {
 	assert.Nil(mt, err, "NewClientEncryption error: %v", err)
 
 	dkOpts := options.DataKey().SetKeyAltNames([]string{keyName})
-	_, err = ce.CreateDataKey(mtest.Background, "local", dkOpts)
+	_, err = ce.CreateDataKey(context.Background(), "local", dkOpts)
 	assert.Nil(mt, err, "CreateDataKey error: %v", err)
 
 	in := bson.RawValue{Type: bsontype.String, Value: bsoncore.AppendString(nil, "test")}
@@ -65,7 +66,7 @@ func createDataKeyAndEncrypt(mt *mtest.T, keyName string) primitive.Binary {
 		SetAlgorithm("AEAD_AES_256_CBC_HMAC_SHA_512-Random").
 		SetKeyAltName(keyName)
 
-	ciphertext, err := ce.Encrypt(mtest.Background, in, eOpts)
+	ciphertext, err := ce.Encrypt(context.Background(), in, eOpts)
 	assert.Nil(mt, err, "Encrypt error: %v", err)
 	return ciphertext
 }
@@ -95,7 +96,6 @@ func makeMonitor(mt *mtest.T, captured *[]event.CommandStartedEvent) *event.Comm
 func TestClientSideEncryptionWithExplicitSessions(t *testing.T) {
 	verifyClientSideEncryptionVarsSet(t)
 	mt := mtest.New(t, mtest.NewOptions().MinServerVersion("4.2").Enterprise(true).CreateClient(false))
-	defer mt.Close()
 
 	kmsProvidersMap := map[string]map[string]interface{}{
 		"local": {"key": localMasterKey},
@@ -121,7 +121,8 @@ func TestClientSideEncryptionWithExplicitSessions(t *testing.T) {
 		aeOpts := options.AutoEncryption().
 			SetKmsProviders(kmsProvidersMap).
 			SetKeyVaultNamespace("keyvault.datakeys").
-			SetSchemaMap(schemaMap)
+			SetSchemaMap(schemaMap).
+			SetExtraOptions(getCryptSharedLibExtraOptions())
 
 		var capturedEvents []event.CommandStartedEvent
 
@@ -132,19 +133,19 @@ func TestClientSideEncryptionWithExplicitSessions(t *testing.T) {
 			SetAutoEncryptionOptions(aeOpts).
 			SetMonitor(makeMonitor(mt, &capturedEvents))
 
-		testutil.AddTestServerAPIVersion(clientOpts)
+		integtest.AddTestServerAPIVersion(clientOpts)
 
-		client, err := mongo.Connect(mtest.Background, clientOpts)
+		client, err := mongo.Connect(context.Background(), clientOpts)
 		assert.Nil(mt, err, "Connect error: %v", err)
-		defer client.Disconnect(mtest.Background)
+		defer client.Disconnect(context.Background())
 
 		coll := client.Database("db").Collection("coll")
-		err = coll.Drop(mtest.Background)
+		err = coll.Drop(context.Background())
 		assert.Nil(mt, err, "Drop error: %v", err)
 
 		session, err := client.StartSession()
 		assert.Nil(mt, err, "StartSession error: %v", err)
-		sessionCtx := mongo.NewSessionContext(mtest.Background, session)
+		sessionCtx := mongo.NewSessionContext(context.Background(), session)
 
 		capturedEvents = make([]event.CommandStartedEvent, 0)
 		_, err = coll.InsertOne(sessionCtx, bson.D{{"encryptMe", "test"}, {"keyName", "myKey"}})
@@ -194,21 +195,21 @@ func TestClientSideEncryptionWithExplicitSessions(t *testing.T) {
 			SetAutoEncryptionOptions(aeOpts).
 			SetMonitor(makeMonitor(mt, &capturedEvents))
 
-		testutil.AddTestServerAPIVersion(clientOpts)
+		integtest.AddTestServerAPIVersion(clientOpts)
 
-		client, err := mongo.Connect(mtest.Background, clientOpts)
+		client, err := mongo.Connect(context.Background(), clientOpts)
 		assert.Nil(mt, err, "Connect error: %v", err)
-		defer client.Disconnect(mtest.Background)
+		defer client.Disconnect(context.Background())
 
 		coll := client.Database("db").Collection("coll")
-		err = coll.Drop(mtest.Background)
+		err = coll.Drop(context.Background())
 		assert.Nil(mt, err, "Drop error: %v", err)
-		_, err = coll.InsertOne(mtest.Background, bson.D{{"encryptMe", ciphertext}})
+		_, err = coll.InsertOne(context.Background(), bson.D{{"encryptMe", ciphertext}})
 		assert.Nil(mt, err, "InsertOne error: %v", err)
 
 		session, err := client.StartSession()
 		assert.Nil(mt, err, "StartSession error: %v", err)
-		sessionCtx := mongo.NewSessionContext(mtest.Background, session)
+		sessionCtx := mongo.NewSessionContext(context.Background(), session)
 
 		capturedEvents = make([]event.CommandStartedEvent, 0)
 		res := coll.FindOne(sessionCtx, bson.D{{}})
@@ -241,13 +242,15 @@ func TestClientSideEncryptionWithExplicitSessions(t *testing.T) {
 // customCrypt is a test implementation of the driver.Crypt interface. It keeps track of the number of times its
 // methods have been called.
 type customCrypt struct {
-	numEncryptCalls              int
-	numDecryptCalls              int
-	numCreateDataKeyCalls        int
-	numEncryptExplicitCalls      int
-	numDecryptExplicitCalls      int
-	numCloseCalls                int
-	numBypassAutoEncryptionCalls int
+	numEncryptCalls                   int
+	numDecryptCalls                   int
+	numCreateDataKeyCalls             int
+	numEncryptExplicitCalls           int
+	numEncryptExplicitExpressionCalls int
+	numDecryptExplicitCalls           int
+	numCloseCalls                     int
+	numBypassAutoEncryptionCalls      int
+	numRewrapDataKeyCalls             int
 }
 
 var (
@@ -307,6 +310,12 @@ func (c *customCrypt) EncryptExplicit(_ context.Context, _ bsoncore.Value, _ *mc
 	return 0, nil, nil
 }
 
+// EncryptExplicit implements the driver.Crypt interface.
+func (c *customCrypt) EncryptExplicitExpression(_ context.Context, _ bsoncore.Document, _ *mcopts.ExplicitEncryptionOptions) (bsoncore.Document, error) {
+	c.numEncryptExplicitExpressionCalls++
+	return nil, nil
+}
+
 // DecryptExplicit implements the driver.Crypt interface.
 func (c *customCrypt) DecryptExplicit(_ context.Context, _ byte, _ []byte) (bsoncore.Value, error) {
 	c.numDecryptExplicitCalls++
@@ -324,9 +333,17 @@ func (c *customCrypt) BypassAutoEncryption() bool {
 	return false
 }
 
+// RewrapDataKey attempts to rewrap the document data keys matching the filter, preparing the re-wrapped documents to
+// be returned as a slice of bsoncore.Document.
+func (c *customCrypt) RewrapDataKey(_ context.Context, _ []byte,
+	_ *mcopts.RewrapManyDataKeyOptions) ([]bsoncore.Document, error) {
+
+	c.numRewrapDataKeyCalls++
+	return nil, nil
+}
+
 func TestClientSideEncryptionCustomCrypt(t *testing.T) {
 	mt := mtest.New(t, mtest.NewOptions().MinServerVersion("4.2").Enterprise(true).CreateClient(false))
-	defer mt.Close()
 
 	kmsProvidersMap := map[string]map[string]interface{}{
 		"local": {"key": localMasterKey},
@@ -335,30 +352,31 @@ func TestClientSideEncryptionCustomCrypt(t *testing.T) {
 	mt.Run("auto encryption and decryption", func(mt *mtest.T) {
 		aeOpts := options.AutoEncryption().
 			SetKmsProviders(kmsProvidersMap).
-			SetKeyVaultNamespace("keyvault.datakeys")
+			SetKeyVaultNamespace("keyvault.datakeys").
+			SetExtraOptions(getCryptSharedLibExtraOptions())
 		clientOpts := options.Client().
 			ApplyURI(mtest.ClusterURI()).
 			SetAutoEncryptionOptions(aeOpts)
 		cc := &customCrypt{}
 		clientOpts.Crypt = cc
-		testutil.AddTestServerAPIVersion(clientOpts)
+		integtest.AddTestServerAPIVersion(clientOpts)
 
-		client, err := mongo.Connect(mtest.Background, clientOpts)
-		defer client.Disconnect(mtest.Background)
+		client, err := mongo.Connect(context.Background(), clientOpts)
+		defer client.Disconnect(context.Background())
 		assert.Nil(mt, err, "Connect error: %v", err)
 
 		coll := client.Database("db").Collection("coll")
-		defer func() { _ = coll.Drop(mtest.Background) }()
+		defer func() { _ = coll.Drop(context.Background()) }()
 
 		doc := bson.D{{"foo", "bar"}, {"ssn", mySSN}}
-		_, err = coll.InsertOne(mtest.Background, doc)
+		_, err = coll.InsertOne(context.Background(), doc)
 		assert.Nil(mt, err, "InsertOne error: %v", err)
 
-		res := coll.FindOne(mtest.Background, bson.D{{"foo", "bar"}})
+		res := coll.FindOne(context.Background(), bson.D{{"foo", "bar"}})
 		assert.Nil(mt, res.Err(), "FindOne error: %v", err)
 
-		rawRes, err := res.DecodeBytes()
-		assert.Nil(mt, err, "DecodeBytes error: %v", err)
+		rawRes, err := res.Raw()
+		assert.Nil(mt, err, "Raw error: %v", err)
 		ssn, ok := rawRes.Lookup("ssn").StringValueOK()
 		assert.True(mt, ok, "expected 'ssn' value to be type string, got %T", ssn)
 		assert.Equal(mt, ssn, mySSN, "expected 'ssn' value %q, got %q", mySSN, ssn)
@@ -372,6 +390,8 @@ func TestClientSideEncryptionCustomCrypt(t *testing.T) {
 			"expected 0 calls to CreateDataKey, got %v", cc.numCreateDataKeyCalls)
 		assert.Equal(mt, cc.numEncryptExplicitCalls, 0,
 			"expected 0 calls to EncryptExplicit, got %v", cc.numEncryptExplicitCalls)
+		assert.Equal(mt, cc.numEncryptExplicitExpressionCalls, 0,
+			"expected 0 calls to EncryptExplicitExpression, got %v", cc.numEncryptExplicitExpressionCalls)
 		assert.Equal(mt, cc.numDecryptExplicitCalls, 0,
 			"expected 0 calls to DecryptExplicit, got %v", cc.numDecryptExplicitCalls)
 		assert.Equal(mt, cc.numCloseCalls, 0,
@@ -380,3 +400,200 @@ func TestClientSideEncryptionCustomCrypt(t *testing.T) {
 			"expected 2 calls to BypassAutoEncryption, got %v", cc.numBypassAutoEncryptionCalls)
 	})
 }
+
+func TestFLE2CreateCollection(t *testing.T) {
+	// FLE 2 (aka Queryable Encryption) is not supported on Standalone topology.
+	mtOpts := mtest.NewOptions().
+		MinServerVersion("7.0").
+		Enterprise(true).
+		CreateClient(false).
+		Topologies(mtest.ReplicaSet,
+			mtest.Sharded,
+			mtest.LoadBalanced,
+			mtest.ShardedReplicaSet)
+	mt := mtest.New(t, mtOpts)
+
+	efJSON := `
+	{
+		"escCollection": "enxcol_.encryptedCollection.esc",
+		"ecocCollection": "enxcol_.encryptedCollection.ecoc",
+		"fields": [
+		  {
+			"path": "firstName",
+			"bsonType": "string",
+			"keyId": {
+			  "$binary": {
+				"subType": "04",
+				"base64": "AAAAAAAAAAAAAAAAAAAAAA=="
+			  }
+			}
+		  }
+		]
+	  }
+	`
+	var efBSON bson.Raw
+	err := bson.UnmarshalExtJSON([]byte(efJSON), true /* canonical */, &efBSON)
+	assert.Nil(mt, err, "UnmarshalExtJSON error: %v", err)
+
+	// Test the behavior in the specification test fle2-CreateCollection.json: "CreateCollection from encryptedFields.".
+	// The Go driver does not support encryptedFields as an option to Drop. See: GODRIVER-2413.
+	mt.Run("CreateCollection from encryptedFields", func(mt *mtest.T) {
+		// Drop data and state collections to clean up from a prior test run.
+		{
+			err = mt.DB.Collection("enxcol_.encryptedCollection.esc").Drop(context.Background())
+			assert.Nil(mt, err, "error in Drop: %v", err)
+			err = mt.DB.Collection("enxcol_.encryptedCollection.ecoc").Drop(context.Background())
+			assert.Nil(mt, err, "error in Drop: %v", err)
+			err := mt.DB.Collection("coll").Drop(context.Background())
+			assert.Nil(mt, err, "error in Drop: %v", err)
+		}
+
+		mt.DB.CreateCollection(context.Background(), "coll", options.CreateCollection().SetEncryptedFields(efBSON))
+
+		// Check expected collections and index exist.
+		{
+			got, err := mt.DB.ListCollectionNames(context.Background(), bson.D{{"name", "coll"}})
+			assert.Nil(mt, err, "error in ListCollectionNames")
+			assert.Equal(mt, got, []string{"coll"}, "expected ['coll'], got: %v", got)
+
+			got, err = mt.DB.ListCollectionNames(context.Background(), bson.D{{"name", "enxcol_.encryptedCollection.esc"}})
+			assert.Nil(mt, err, "error in ListCollectionNames")
+			assert.Equal(mt, got, []string{"enxcol_.encryptedCollection.esc"}, "expected ['encryptedCollection.esc'], got: %v", got)
+
+			got, err = mt.DB.ListCollectionNames(context.Background(), bson.D{{"name", "enxcol_.encryptedCollection.ecoc"}})
+			assert.Nil(mt, err, "error in ListCollectionNames")
+			assert.Equal(mt, got, []string{"enxcol_.encryptedCollection.ecoc"}, "expected ['encryptedCollection.ecoc'], got: %v", got)
+
+			indexSpecs, err := mt.DB.Collection("coll").Indexes().ListSpecifications(context.Background())
+			assert.Nil(mt, err, "error in Indexes().ListSpecifications: %v", err)
+			assert.Equal(mt, len(indexSpecs), 2, "expected two indexes on 'coll', got: %v", indexSpecs)
+			assert.Equal(mt, indexSpecs[1].Name, "__safeContent___1", "expected second index to be '__safeContent___1', got %v", indexSpecs[1].Name)
+		}
+	})
+}
+
+func TestFLE2DocsExample(t *testing.T) {
+	// FLE 2 is not supported on Standalone topology.
+	// Only test MongoDB Server 7.0+. MongoDB Server 7.0 introduced a backwards breaking change to the Queryable Encryption (QE) protocol: QEv2.
+	// libmongocrypt is configured to use the QEv2 protocol.
+	mtOpts := mtest.NewOptions().
+		MinServerVersion("7.0").
+		Enterprise(true).
+		CreateClient(false).
+		Topologies(mtest.ReplicaSet,
+			mtest.Sharded,
+			mtest.LoadBalanced,
+			mtest.ShardedReplicaSet)
+	mt := mtest.New(t, mtOpts)
+
+	mt.Run("Auto Encryption", func(mt *mtest.T) {
+		// Drop data from prior test runs.
+		{
+			err := mt.Client.Database("keyvault").Collection("datakeys").Drop(context.Background())
+			assert.Nil(mt, err, "error in Drop: %v", err)
+			err = mt.Client.Database("docsExamples").Drop(context.Background())
+			assert.Nil(mt, err, "error in Drop: %v", err)
+		}
+
+		kmsProvidersMap := map[string]map[string]interface{}{
+			"local": {"key": localMasterKey},
+		}
+
+		var key1ID primitive.Binary
+		var key2ID primitive.Binary
+
+		// Create two data keys.
+		{
+			cOpts := options.Client().ApplyURI(mtest.ClusterURI())
+			integtest.AddTestServerAPIVersion(cOpts)
+			keyVaultClient, err := mongo.Connect(context.Background(), cOpts)
+			assert.Nil(mt, err, "error in Connect: %v", err)
+			defer keyVaultClient.Disconnect(context.Background())
+			ceOpts := options.ClientEncryption().SetKmsProviders(kmsProvidersMap).SetKeyVaultNamespace("keyvault.datakeys")
+			ce, err := mongo.NewClientEncryption(keyVaultClient, ceOpts)
+			assert.Nil(mt, err, "error in NewClientEncryption: %v", err)
+			defer ce.Close(context.Background())
+			key1ID, err = ce.CreateDataKey(context.Background(), "local")
+			assert.Nil(mt, err, "error in CreateDataKey: %v", err)
+			key2ID, err = ce.CreateDataKey(context.Background(), "local")
+			assert.Nil(mt, err, "error in CreateDataKey: %v", err)
+		}
+
+		// Create an encryptedFieldsMap.
+		encryptedFieldsMap := bson.M{
+			"docsExamples.encrypted": bson.M{
+				"fields": []bson.M{
+					{
+						"path":     "encryptedIndexed",
+						"bsonType": "string",
+						"keyId":    key1ID,
+						"queries": []bson.M{
+							{
+								"queryType": "equality",
+							},
+						},
+					},
+					{
+						"path":     "encryptedUnindexed",
+						"bsonType": "string",
+						"keyId":    key2ID,
+					},
+				},
+			},
+		}
+
+		// Create an FLE 2 collection.
+		var encryptedColl *mongo.Collection
+		{
+			cOpts := options.Client().ApplyURI(mtest.ClusterURI())
+			integtest.AddTestServerAPIVersion(cOpts)
+			aeOpts := options.AutoEncryption().SetKmsProviders(kmsProvidersMap).SetKeyVaultNamespace("keyvault.datakeys").SetEncryptedFieldsMap(encryptedFieldsMap).SetExtraOptions(getCryptSharedLibExtraOptions())
+			cOpts.SetAutoEncryptionOptions(aeOpts)
+			encryptedClient, err := mongo.Connect(context.Background(), cOpts)
+			defer encryptedClient.Disconnect(context.Background())
+			assert.Nil(mt, err, "error in Connect: %v", err)
+			// Create the FLE 2 collection docsExample.encrypted.
+			db := encryptedClient.Database("docsExamples")
+			// Because docsExample.encrypted is in encryptedFieldsMap, it is created with FLE 2 support.
+			err = db.CreateCollection(context.Background(), "encrypted")
+			assert.Nil(mt, err, "error in CreateCollection")
+			encryptedColl = db.Collection("encrypted")
+		}
+
+		// Auto encrypt an insert and find.
+		{
+			// Encrypt an insert.
+			_, err := encryptedColl.InsertOne(context.Background(), bson.M{
+				"_id":                1,
+				"encryptedIndexed":   "indexedValue",
+				"encryptedUnindexed": "unindexedValue",
+			})
+			assert.Nil(mt, err, "error in InsertOne")
+
+			// Encrypt a find.
+			res := encryptedColl.FindOne(context.Background(), bson.M{
+				"encryptedIndexed": "indexedValue",
+			})
+			assert.Nil(mt, res.Err(), "error in FindOne: %v", res.Err())
+			var resBSON bson.M
+			err = res.Decode(&resBSON)
+			assert.Nil(mt, err, "error in Decode: %v", err)
+			assert.Equal(mt, resBSON["encryptedIndexed"], "indexedValue", "expected 'indexedValue', got %q", resBSON["encryptedIndexed"])
+			assert.Equal(mt, resBSON["encryptedUnindexed"], "unindexedValue", "expected 'unindexedValue', got %q", resBSON["encryptedUnindexed"])
+		}
+
+		// Find documents without decryption.
+		{
+			unencryptedColl := mt.Client.Database("docsExamples").Collection("encrypted")
+			res := unencryptedColl.FindOne(context.Background(), bson.M{"_id": 1})
+			assert.Nil(mt, res.Err(), "error in FindOne: %v", res.Err())
+			resBSON, err := res.Raw()
+			assert.Nil(mt, err, "error in Raw: %v", err)
+
+			val := resBSON.Lookup("encryptedIndexed")
+			assert.Equal(mt, val.Type, bsontype.Binary, "expected encryptedIndexed to be Binary, got %v", val.Type)
+			val = resBSON.Lookup("encryptedUnindexed")
+			assert.Equal(mt, val.Type, bsontype.Binary, "expected encryptedUnindexed to be Binary, got %v", val.Type)
+		}
+	})
+}
