@@ -9,7 +9,9 @@ package mongo
 import (
 	"context"
 	"errors"
+	"fmt"
 	"math"
+	"os"
 	"strconv"
 	"strings"
 	"testing"
@@ -18,8 +20,7 @@ import (
 	"go.mongodb.org/mongo-driver/bson"
 	"go.mongodb.org/mongo-driver/event"
 	"go.mongodb.org/mongo-driver/internal/assert"
-	"go.mongodb.org/mongo-driver/internal/testutil"
-	"go.mongodb.org/mongo-driver/internal/testutil/helpers"
+	"go.mongodb.org/mongo-driver/internal/integtest"
 	"go.mongodb.org/mongo-driver/mongo/description"
 	"go.mongodb.org/mongo-driver/mongo/options"
 	"go.mongodb.org/mongo-driver/mongo/readpref"
@@ -33,22 +34,13 @@ var (
 	errorInterrupted int32 = 11601
 )
 
-type wrappedError struct {
-	err error
-}
-
-func (we wrappedError) Error() string {
-	return we.err.Error()
-}
-
-func (we wrappedError) Unwrap() error {
-	return we.err
-}
-
 func TestConvenientTransactions(t *testing.T) {
 	if testing.Short() {
 		t.Skip("skipping integration test in short mode")
 	}
+	if os.Getenv("DOCKER_RUNNING") != "" {
+		t.Skip("skipping test in docker environment")
+	}
 
 	client := setupConvenientTransactions(t)
 	db := client.Database("TestConvenientTransactions")
@@ -128,7 +120,7 @@ func TestConvenientTransactions(t *testing.T) {
 				"expected error with label %v, got %v", driver.TransientTransactionError, cmdErr)
 		})
 		t.Run("unknown transaction commit result", func(t *testing.T) {
-			//set failpoint
+			// set failpoint
 			failpoint := bson.D{{"configureFailPoint", "failCommand"},
 				{"mode", "alwaysOn"},
 				{"data", bson.D{
@@ -161,7 +153,7 @@ func TestConvenientTransactions(t *testing.T) {
 				"expected error with label %v, got %v", driver.UnknownTransactionCommitResult, cmdErr)
 		})
 		t.Run("commit transient transaction error", func(t *testing.T) {
-			//set failpoint
+			// set failpoint
 			failpoint := bson.D{{"configureFailPoint", "failCommand"},
 				{"mode", "alwaysOn"},
 				{"data", bson.D{
@@ -279,7 +271,7 @@ func TestConvenientTransactions(t *testing.T) {
 		var abortSucceeded []*event.CommandSucceededEvent
 		var abortFailed []*event.CommandFailedEvent
 		monitor := &event.CommandMonitor{
-			Started: func(ctx context.Context, evt *event.CommandStartedEvent) {
+			Started: func(_ context.Context, evt *event.CommandStartedEvent) {
 				if evt.CommandName == "abortTransaction" {
 					abortStarted = append(abortStarted, evt)
 				}
@@ -357,7 +349,7 @@ func TestConvenientTransactions(t *testing.T) {
 		var abortSucceeded []*event.CommandSucceededEvent
 		var abortFailed []*event.CommandFailedEvent
 		monitor := &event.CommandMonitor{
-			Started: func(ctx context.Context, evt *event.CommandStartedEvent) {
+			Started: func(_ context.Context, evt *event.CommandStartedEvent) {
 				if evt.CommandName == "abortTransaction" {
 					abortStarted = append(abortStarted, evt)
 				}
@@ -407,19 +399,23 @@ func TestConvenientTransactions(t *testing.T) {
 
 		// Insert a document within a session and manually cancel context before
 		// "commitTransaction" can be sent.
-		callback := func(ctx context.Context) {
-			transactionCtx, cancel := context.WithCancel(ctx)
-
+		callback := func() bool {
+			transactionCtx, cancel := context.WithCancel(context.Background())
 			_, _ = sess.WithTransaction(transactionCtx, func(ctx SessionContext) (interface{}, error) {
 				_, err := coll.InsertOne(ctx, bson.M{"x": 1})
-				assert.Nil(t, err, "InsertOne error: %v", err)
+				assert.NoError(t, err, "InsertOne error: %v", err)
 				cancel()
 				return nil, nil
 			})
+			return true
 		}
 
 		// Assert that transaction is canceled within 500ms and not 2 seconds.
-		helpers.AssertSoon(t, callback, 500*time.Millisecond)
+		assert.Eventually(t,
+			callback,
+			500*time.Millisecond,
+			time.Millisecond,
+			"expected transaction to be canceled within 500ms")
 
 		// Assert that AbortTransaction was started once and succeeded.
 		assert.Equal(t, 1, len(abortStarted), "expected 1 abortTransaction started event, got %d", len(abortStarted))
@@ -437,12 +433,12 @@ func TestConvenientTransactions(t *testing.T) {
 		res, err := sess.WithTransaction(context.Background(), func(SessionContext) (interface{}, error) {
 			if returnError {
 				returnError = false
-				return nil, wrappedError{
+				return nil, fmt.Errorf("%w",
 					CommandError{
 						Name:   "test Error",
 						Labels: []string{driver.TransientTransactionError},
 					},
-				}
+				)
 			}
 			return false, nil
 		})
@@ -467,19 +463,24 @@ func TestConvenientTransactions(t *testing.T) {
 		assert.Nil(t, err, "StartSession error: %v", err)
 		defer sess.EndSession(context.Background())
 
-		callback := func(ctx context.Context) {
+		callback := func() bool {
 			// Create transaction context with short timeout.
-			withTransactionContext, cancel := context.WithTimeout(ctx, time.Nanosecond)
+			withTransactionContext, cancel := context.WithTimeout(context.Background(), time.Nanosecond)
 			defer cancel()
 
 			_, _ = sess.WithTransaction(withTransactionContext, func(ctx SessionContext) (interface{}, error) {
 				_, err := coll.InsertOne(ctx, bson.D{{}})
 				return nil, err
 			})
+			return true
 		}
 
 		// Assert that transaction fails within 500ms and not 2 seconds.
-		helpers.AssertSoon(t, callback, 500*time.Millisecond)
+		assert.Eventually(t,
+			callback,
+			500*time.Millisecond,
+			time.Millisecond,
+			"expected transaction to fail within 500ms")
 	})
 	t.Run("canceled context before callback does not retry", func(t *testing.T) {
 		withTransactionTimeout = 2 * time.Second
@@ -497,19 +498,24 @@ func TestConvenientTransactions(t *testing.T) {
 		assert.Nil(t, err, "StartSession error: %v", err)
 		defer sess.EndSession(context.Background())
 
-		callback := func(ctx context.Context) {
+		callback := func() bool {
 			// Create transaction context and cancel it immediately.
-			withTransactionContext, cancel := context.WithTimeout(ctx, 2*time.Second)
+			withTransactionContext, cancel := context.WithTimeout(context.Background(), 2*time.Second)
 			cancel()
 
 			_, _ = sess.WithTransaction(withTransactionContext, func(ctx SessionContext) (interface{}, error) {
 				_, err := coll.InsertOne(ctx, bson.D{{}})
 				return nil, err
 			})
+			return true
 		}
 
 		// Assert that transaction fails within 500ms and not 2 seconds.
-		helpers.AssertSoon(t, callback, 500*time.Millisecond)
+		assert.Eventually(t,
+			callback,
+			500*time.Millisecond,
+			time.Millisecond,
+			"expected transaction to fail within 500ms")
 	})
 	t.Run("slow operation in callback retries", func(t *testing.T) {
 		withTransactionTimeout = 2 * time.Second
@@ -548,8 +554,8 @@ func TestConvenientTransactions(t *testing.T) {
 		assert.Nil(t, err, "StartSession error: %v", err)
 		defer sess.EndSession(context.Background())
 
-		callback := func(ctx context.Context) {
-			_, err = sess.WithTransaction(ctx, func(ctx SessionContext) (interface{}, error) {
+		callback := func() bool {
+			_, err = sess.WithTransaction(context.Background(), func(ctx SessionContext) (interface{}, error) {
 				// Set a timeout of 300ms to cause a timeout on first insertOne
 				// and force a retry.
 				c, cancel := context.WithTimeout(ctx, 300*time.Millisecond)
@@ -558,16 +564,22 @@ func TestConvenientTransactions(t *testing.T) {
 				_, err := coll.InsertOne(c, bson.D{{}})
 				return nil, err
 			})
-			assert.Nil(t, err, "WithTransaction error: %v", err)
+			assert.NoError(t, err, "WithTransaction error: %v", err)
+			return true
 		}
 
 		// Assert that transaction passes within 2 seconds.
-		helpers.AssertSoon(t, callback, 2*time.Second)
+		assert.Eventually(t,
+			callback,
+			withTransactionTimeout,
+			time.Millisecond,
+			"expected transaction to be passed within 2s")
+
 	})
 }
 
 func setupConvenientTransactions(t *testing.T, extraClientOpts ...*options.ClientOptions) *Client {
-	cs := testutil.ConnString(t)
+	cs := integtest.ConnString(t)
 	poolMonitor := &event.PoolMonitor{
 		Event: func(evt *event.PoolEvent) {
 			switch evt.Type {
@@ -584,7 +596,7 @@ func setupConvenientTransactions(t *testing.T, extraClientOpts ...*options.Clien
 		SetReadPreference(readpref.Primary()).
 		SetWriteConcern(writeconcern.New(writeconcern.WMajority())).
 		SetPoolMonitor(poolMonitor)
-	testutil.AddTestServerAPIVersion(baseClientOpts)
+	integtest.AddTestServerAPIVersion(baseClientOpts)
 	fullClientOpts := []*options.ClientOptions{baseClientOpts}
 	fullClientOpts = append(fullClientOpts, extraClientOpts...)
 
@@ -614,7 +626,7 @@ func getServerVersion(db *Database) (string, error) {
 	serverStatus, err := db.RunCommand(
 		context.Background(),
 		bson.D{{"serverStatus", 1}},
-	).DecodeBytes()
+	).Raw()
 	if err != nil {
 		return "", err
 	}
