@@ -14,7 +14,7 @@ import (
 	"testing"
 
 	"go.mongodb.org/mongo-driver/internal/require"
-	"go.mongodb.org/mongo-driver/internal/testutil/helpers"
+	"go.mongodb.org/mongo-driver/internal/spectest"
 	"go.mongodb.org/mongo-driver/mongo/options"
 )
 
@@ -105,7 +105,7 @@ func mapInterfaceToString(m map[string]interface{}) map[string]string {
 
 // Test case for all connection string spec tests.
 func TestAuthSpec(t *testing.T) {
-	for _, file := range helpers.FindJSONFilesInDir(t, authTestsDir) {
+	for _, file := range spectest.FindJSONFilesInDir(t, authTestsDir) {
 		runTestsInFile(t, authTestsDir, file)
 	}
 }
