@@ -8,6 +8,7 @@ package auth
 
 import (
 	"context"
+	"net/http"
 	"testing"
 
 	"go.mongodb.org/mongo-driver/internal/assert"
@@ -38,7 +39,7 @@ func TestSCRAM(t *testing.T) {
 	t.Run("conversation", func(t *testing.T) {
 		testCases := []struct {
 			name                  string
-			createAuthenticatorFn func(*Cred) (Authenticator, error)
+			createAuthenticatorFn func(*Cred, *http.Client) (Authenticator, error)
 			payloads              [][]byte
 			nonce                 string
 		}{
@@ -49,11 +50,13 @@ func TestSCRAM(t *testing.T) {
 		}
 		for _, tc := range testCases {
 			t.Run(tc.name, func(t *testing.T) {
-				authenticator, err := tc.createAuthenticatorFn(&Cred{
-					Username: "user",
-					Password: "pencil",
-					Source:   "admin",
-				})
+				authenticator, err := tc.createAuthenticatorFn(
+					&Cred{
+						Username: "user",
+						Password: "pencil",
+						Source:   "admin",
+					},
+					&http.Client{})
 				assert.Nil(t, err, "error creating authenticator: %v", err)
 				sa, _ := authenticator.(*ScramAuthenticator)
 				sa.client = sa.client.WithNonceGenerator(func() string {
@@ -65,7 +68,7 @@ func TestSCRAM(t *testing.T) {
 
 				desc := description.Server{
 					WireVersion: &description.VersionRange{
-						Max: 4,
+						Max: 21,
 					},
 				}
 				conn := &drivertest.ChannelConn{
@@ -79,7 +82,7 @@ func TestSCRAM(t *testing.T) {
 
 				// Verify that the first command sent is saslStart.
 				assert.True(t, len(conn.Written) > 1, "wire messages were written to the connection")
-				startCmd, err := drivertest.GetCommandFromQueryWireMessage(<-conn.Written)
+				startCmd, err := drivertest.GetCommandFromMsgWireMessage(<-conn.Written)
 				assert.Nil(t, err, "error parsing wire message: %v", err)
 				cmdName := startCmd.Index(0).Key()
 				assert.Equal(t, cmdName, "saslStart", "cmd name mismatch; expected 'saslStart', got %v", cmdName)
