@@ -10,22 +10,31 @@ import (
 	"context"
 	"errors"
 	"fmt"
+	"io"
 	"strings"
+	"time"
 
+	"go.mongodb.org/mongo-driver/bson"
 	"go.mongodb.org/mongo-driver/bson/bsontype"
 	"go.mongodb.org/mongo-driver/event"
-	"go.mongodb.org/mongo-driver/internal"
+	"go.mongodb.org/mongo-driver/internal/codecutil"
+	"go.mongodb.org/mongo-driver/internal/csot"
 	"go.mongodb.org/mongo-driver/mongo/description"
 	"go.mongodb.org/mongo-driver/x/bsonx/bsoncore"
 	"go.mongodb.org/mongo-driver/x/mongo/driver/session"
 )
 
+// ErrNoCursor is returned by NewCursorResponse when the database response does
+// not contain a cursor.
+var ErrNoCursor = errors.New("database response does not contain a cursor")
+
 // BatchCursor is a batch implementation of a cursor. It returns documents in entire batches instead
 // of one at a time. An individual document cursor can be built on top of this batch cursor.
 type BatchCursor struct {
 	clientSession        *session.Client
 	clock                *session.ClusterClock
-	comment              bsoncore.Value
+	comment              interface{}
+	encoderFn            codecutil.EncoderFn
 	database             string
 	collection           string
 	id                   int64
@@ -62,17 +71,27 @@ type CursorResponse struct {
 	postBatchResumeToken bsoncore.Document
 }
 
-// NewCursorResponse constructs a cursor response from the given response and server. This method
-// can be used within the ProcessResponse method for an operation.
+// NewCursorResponse constructs a cursor response from the given response and
+// server. If the provided database response does not contain a cursor, it
+// returns ErrNoCursor.
+//
+// NewCursorResponse can be used within the ProcessResponse method for an operation.
 func NewCursorResponse(info ResponseInfo) (CursorResponse, error) {
 	response := info.ServerResponse
-	cur, ok := response.Lookup("cursor").DocumentOK()
-	if !ok {
-		return CursorResponse{}, fmt.Errorf("cursor should be an embedded document but is of BSON type %s", response.Lookup("cursor").Type)
+	cur, err := response.LookupErr("cursor")
+	if errors.Is(err, bsoncore.ErrElementNotFound) {
+		return CursorResponse{}, ErrNoCursor
 	}
-	elems, err := cur.Elements()
 	if err != nil {
-		return CursorResponse{}, err
+		return CursorResponse{}, fmt.Errorf("error getting cursor from database response: %w", err)
+	}
+	curDoc, ok := cur.DocumentOK()
+	if !ok {
+		return CursorResponse{}, fmt.Errorf("cursor should be an embedded document but is BSON type %s", cur.Type)
+	}
+	elems, err := curDoc.Elements()
+	if err != nil {
+		return CursorResponse{}, fmt.Errorf("error getting elements from cursor: %w", err)
 	}
 	curresp := CursorResponse{Server: info.Server, Desc: info.ConnectionDescription}
 
@@ -89,12 +108,12 @@ func NewCursorResponse(info ResponseInfo) (CursorResponse, error) {
 			if !ok {
 				return CursorResponse{}, fmt.Errorf("ns should be a string but is a BSON %s", elem.Value().Type)
 			}
-			index := strings.Index(ns, ".")
-			if index == -1 {
+			database, collection, ok := strings.Cut(ns, ".")
+			if !ok {
 				return CursorResponse{}, errors.New("ns field must contain a valid namespace, but is missing '.'")
 			}
-			curresp.Database = ns[:index]
-			curresp.Collection = ns[index+1:]
+			curresp.Database = database
+			curresp.Collection = collection
 		case "id":
 			curresp.ID, ok = elem.Value().Int64OK()
 			if !ok {
@@ -123,7 +142,7 @@ func NewCursorResponse(info ResponseInfo) (CursorResponse, error) {
 			return CursorResponse{}, fmt.Errorf("expected Connection used to establish a cursor to implement PinnedConnection, but got %T", info.Connection)
 		}
 		if err := refConn.PinToCursor(); err != nil {
-			return CursorResponse{}, fmt.Errorf("error incrementing connection reference count when creating a cursor: %v", err)
+			return CursorResponse{}, fmt.Errorf("error incrementing connection reference count when creating a cursor: %w", err)
 		}
 		curresp.Connection = refConn
 	}
@@ -133,13 +152,14 @@ func NewCursorResponse(info ResponseInfo) (CursorResponse, error) {
 
 // CursorOptions are extra options that are required to construct a BatchCursor.
 type CursorOptions struct {
-	BatchSize      int32
-	Comment        bsoncore.Value
-	MaxTimeMS      int64
-	Limit          int32
-	CommandMonitor *event.CommandMonitor
-	Crypt          Crypt
-	ServerAPI      *ServerAPIOptions
+	BatchSize             int32
+	Comment               bsoncore.Value
+	MaxTimeMS             int64
+	Limit                 int32
+	CommandMonitor        *event.CommandMonitor
+	Crypt                 Crypt
+	ServerAPI             *ServerAPIOptions
+	MarshalValueEncoderFn func(io.Writer) (*bson.Encoder, error)
 }
 
 // NewBatchCursor creates a new BatchCursor from the provided parameters.
@@ -163,12 +183,13 @@ func NewBatchCursor(cr CursorResponse, clientSession *session.Client, clock *ses
 		crypt:                opts.Crypt,
 		serverAPI:            opts.ServerAPI,
 		serverDescription:    cr.Desc,
+		encoderFn:            opts.MarshalValueEncoderFn,
 	}
 
 	if ds != nil {
 		bc.numReturned = int32(ds.DocumentCount())
 	}
-	if cr.Desc.WireVersion == nil || cr.Desc.WireVersion.Max < 4 {
+	if cr.Desc.WireVersion == nil {
 		bc.limit = opts.Limit
 
 		// Take as many documents from the batch as needed.
@@ -293,7 +314,7 @@ func (bc *BatchCursor) KillCursor(ctx context.Context) error {
 	}
 
 	return Operation{
-		CommandFn: func(dst []byte, desc description.SelectedServer) ([]byte, error) {
+		CommandFn: func(dst []byte, _ description.SelectedServer) ([]byte, error) {
 			dst = bsoncore.AppendStringElement(dst, "killCursors", bc.collection)
 			dst = bsoncore.BuildArrayElement(dst, "cursors", bsoncore.Value{Type: bsontype.Int64, Data: bsoncore.AppendInt64(nil, bc.id)})
 			return dst, nil
@@ -305,6 +326,12 @@ func (bc *BatchCursor) KillCursor(ctx context.Context) error {
 		Legacy:         LegacyKillCursors,
 		CommandMonitor: bc.cmdMonitor,
 		ServerAPI:      bc.serverAPI,
+
+		// No read preference is passed to the killCursor command,
+		// resulting in the default read preference: "primaryPreferred".
+		// Since this could be confusing, and there is no requirement
+		// to use a read preference here, we omit it.
+		omitReadPreference: true,
 	}.Execute(ctx)
 }
 
@@ -342,7 +369,7 @@ func (bc *BatchCursor) getMore(ctx context.Context) {
 	}
 
 	bc.err = Operation{
-		CommandFn: func(dst []byte, desc description.SelectedServer) ([]byte, error) {
+		CommandFn: func(dst []byte, _ description.SelectedServer) ([]byte, error) {
 			dst = bsoncore.AppendInt64Element(dst, "getMore", bc.id)
 			dst = bsoncore.AppendStringElement(dst, "collection", bc.collection)
 			if numToReturn > 0 {
@@ -351,10 +378,17 @@ func (bc *BatchCursor) getMore(ctx context.Context) {
 			if bc.maxTimeMS > 0 {
 				dst = bsoncore.AppendInt64Element(dst, "maxTimeMS", bc.maxTimeMS)
 			}
-			// The getMore command does not support commenting pre-4.4.
-			if bc.comment.Type != bsontype.Type(0) && bc.serverDescription.WireVersion.Max >= 9 {
-				dst = bsoncore.AppendValueElement(dst, "comment", bc.comment)
+
+			comment, err := codecutil.MarshalValue(bc.comment, bc.encoderFn)
+			if err != nil {
+				return nil, fmt.Errorf("error marshaling comment as a BSON value: %w", err)
 			}
+
+			// The getMore command does not support commenting pre-4.4.
+			if comment.Type != bsontype.Type(0) && bc.serverDescription.WireVersion.Max >= 9 {
+				dst = bsoncore.AppendValueElement(dst, "comment", comment)
+			}
+
 			return dst, nil
 		},
 		Database:   bc.database,
@@ -398,6 +432,12 @@ func (bc *BatchCursor) getMore(ctx context.Context) {
 		CommandMonitor: bc.cmdMonitor,
 		Crypt:          bc.crypt,
 		ServerAPI:      bc.serverAPI,
+
+		// No read preference is passed to the getMore command,
+		// resulting in the default read preference: "primaryPreferred".
+		// Since this could be confusing, and there is no requirement
+		// to use a read preference here, we omit it.
+		omitReadPreference: true,
 	}.Execute(ctx)
 
 	// Once the cursor has been drained, we can unpin the connection if one is currently pinned.
@@ -430,11 +470,26 @@ func (bc *BatchCursor) PostBatchResumeToken() bsoncore.Document {
 	return bc.postBatchResumeToken
 }
 
-// SetBatchSize sets the batchSize for future getMores.
+// SetBatchSize sets the batchSize for future getMore operations.
 func (bc *BatchCursor) SetBatchSize(size int32) {
 	bc.batchSize = size
 }
 
+// SetMaxTime will set the maximum amount of time the server will allow the
+// operations to execute. The server will error if this field is set but the
+// cursor is not configured with awaitData=true.
+//
+// The time.Duration value passed by this setter will be converted and rounded
+// down to the nearest millisecond.
+func (bc *BatchCursor) SetMaxTime(dur time.Duration) {
+	bc.maxTimeMS = int64(dur / time.Millisecond)
+}
+
+// SetComment sets the comment for future getMore operations.
+func (bc *BatchCursor) SetComment(comment interface{}) {
+	bc.comment = comment
+}
+
 func (bc *BatchCursor) getOperationDeployment() Deployment {
 	if bc.connection != nil {
 		return &loadBalancedCursorDeployment{
@@ -471,7 +526,7 @@ func (lbcd *loadBalancedCursorDeployment) Connection(_ context.Context) (Connect
 
 // RTTMonitor implements the driver.Server interface.
 func (lbcd *loadBalancedCursorDeployment) RTTMonitor() RTTMonitor {
-	return &internal.ZeroRTTMonitor{}
+	return &csot.ZeroRTTMonitor{}
 }
 
 func (lbcd *loadBalancedCursorDeployment) ProcessError(err error, conn Connection) ProcessErrorResult {
