@@ -13,13 +13,12 @@ import (
 
 // ExecuteExhaust reads a response from the provided StreamerConnection. This will error if the connection's
 // CurrentlyStreaming function returns false.
-func (op Operation) ExecuteExhaust(ctx context.Context, conn StreamerConnection, scratch []byte) error {
+func (op Operation) ExecuteExhaust(ctx context.Context, conn StreamerConnection) error {
 	if !conn.CurrentlyStreaming() {
 		return errors.New("exhaust read must be done with a connection that is currently streaming")
 	}
 
-	scratch = scratch[:0]
-	res, err := op.readWireMessage(ctx, conn, scratch)
+	res, err := op.readWireMessage(ctx, conn)
 	if err != nil {
 		return err
 	}
