@@ -105,6 +105,7 @@ func (ss serverSorter) Less(i, j int) bool {
 }
 
 func compareHosts(t *testing.T, received []description.Server, expected []string) {
+	t.Helper()
 	if len(received) != len(expected) {
 		t.Fatalf("Number of hosts in topology does not match expected value. Got %v; want %v.", len(received), len(expected))
 	}
@@ -182,8 +183,12 @@ func TestPollSRVRecords(t *testing.T) {
 		topo.serversLock.Lock()
 		topo.fsm.Kind = description.Single
 		topo.desc.Store(description.Topology{
-			Kind:                  topo.fsm.Kind,
-			Servers:               topo.fsm.Servers,
+			Kind:                     topo.fsm.Kind,
+			Servers:                  topo.fsm.Servers,
+			SessionTimeoutMinutesPtr: topo.fsm.SessionTimeoutMinutesPtr,
+
+			// TODO(GODRIVER-2885): This field can be removed once
+			// legacy SessionTimeoutMinutes is removed.
 			SessionTimeoutMinutes: topo.fsm.SessionTimeoutMinutes,
 		})
 		topo.serversLock.Unlock()
@@ -308,7 +313,7 @@ func TestPollingSRVRecordsLoadBalanced(t *testing.T) {
 
 func TestPollSRVRecordsMaxHosts(t *testing.T) {
 	// simulateSRVPoll creates a topology with srvMaxHosts, mocks the DNS changes described by
-	// recordsToAdd and recordsToRemove, and returns the the topology.
+	// recordsToAdd and recordsToRemove, and returns the topology.
 	simulateSRVPoll := func(srvMaxHosts int, recordsToAdd []*net.SRV, recordsToRemove []*net.SRV) (*Topology, func(ctx context.Context) error) {
 		t.Helper()
 
